# Whole new package, so include pack stuff
require 'puppet/provider/package'

# we need this to allocate a new pty 
require "pty"

Puppet::Type.type(:package).provide :portupgrade_pkgng, :parent => Puppet::Provider::Package do
  include Puppet::Util::Execution

  desc "Support for FreeBSD's ports using the portupgrade ports management software.
    Use the port's full origin as the resource name. eg (ports-mgmt/portupgrade)
    for the portupgrade port."

  ## has_features is usually autodetected based on defs below.
  # has_features :installable, :uninstallable, :upgradeable

  commands :portupgrade   => "/usr/local/sbin/portupgrade",
  ###:portinstall   => "/usr/local/sbin/portinstall",
  :portinstall   => "/usr/local/sbin/portinstall",
#  :portversion   => "/usr/local/sbin/portversion",
#  :portuninstall => "/usr/local/sbin/pkg_deinstall",
#  :portinfo      => "/usr/sbin/pkg_info",
  :pkg           => "/usr/sbin/pkg"

  ## Activate this only once approved by someone important.
  # defaultfor :operatingsystem => :freebsd

  # Remove unwanted environment variables.
  %w{INTERACTIVE UNAME}.each do |var|
    if ENV.include?(var)
      ENV.delete(var)
    end
  end


  ######## instances sub command (builds the installed packages list)

  def self.instances
    Puppet.debug "pkg query -a %n-%v:%o"

    # regex to match output from pkg_info
    regex = %r{^(\S+)-([^-\s]+):(\S+)$}
    # Corresponding field names
    fields = [:portname, :ensure, :portorigin]
    # define Temporary hash used, packages array of hashes
    packages = []

    # exec command
    cmdline = ["query",'-a','%n-%v:%o']
    begin
      output = pkg(*cmdline)
    rescue Puppet::ExecutionFailure
      raise Puppet::Error.new(output)
      return nil
    end

    # split output and match it and populate temp hash
    output.split("\n").each { |data|
      # reset hash to nil for each line
      hash = Hash.new
      if match = regex.match(data)
        # Output matched regex
        fields.zip(match.captures) { |field, value|
          hash[field] = value
          ### Puppet.debug "#{field} = #{value}"
        }

        # populate the actual :name field from the :portorigin
        # Set :provider to this object name
        hash[:name] = hash[:portorigin]
        hash[:provider] = self.name

        # Add to the full packages listing
	### Puppet.debug hash.to_yaml
        packages << new(hash)
      else
        # unrecognised output from pkg_info
        Puppet.debug "portupgrade.Instances() - unable to match output: #{data}"
      end
    }

    Puppet.debug packages.to_yaml
    # return the packages array of hashes
    return packages
  end

  ######## Installation sub command

  def install


    a_pty_device = PTY.open[1].path

    Puppet.debug "portinstall #{@resource[:name]}"
    Puppet.debug "allocated pty device #{a_pty_device}"

    cmdline = ['/usr/local/sbin/portinstall','--batch','-v','-b','-M BATCH=yes', @resource[:name]]

    # using the console as input ensures that the script process spawned by portinstall will not hang
    begin
      options = { :stdinfile => a_pty_device, :combine => :true, :raise => :true, :failonfail => :true }
      output = Puppet::Util::Execution.execute( cmdline, options ) 
    rescue Puppet::ExecutionFailure
      raise Puppet::Error.new(output)
    end


    if output =~ /\*\* No such /
      raise Puppet::ExecutionFailure, "Could not find package #{@resource[:name]}"
    end

    Puppet.debug output
    # No return code required, so do nil to be clean
    return nil
  end

  ######## Latest subcommand (returns the latest version available, or current version if installed is latest)

  def latest
    Puppet.debug "pkg version -ve Latest check called on #{@resource[:name]}"
    # search for latest version available, or return current version.
    # cmdline = "portversion -v <portorigin>", returns "<portname> <code> <stuff>"
    # or "** No matching package found: <portname>"
    cmdline = ['version','-ve', @resource[:name]]

    begin
      output = pkg(*cmdline)
    rescue Puppet::ExecutionFailure
      raise Puppet::Error.new(output)
    end

    # Check: output format.
    if output =~ /^\S+-([^-\s]+)\s+(\S)\s+(.*)/
      installedversion = $1
      comparison = $2
      otherdata = $3

      # Only return a new version number when it's clear that there is a new version
      # all others return the current version so no unexpected 'upgrades' occur.
      case comparison
      # screen-4.2.1_1                     =   up-to-date with port
      when "=", ">"
        Puppet.debug "pkg version  Installed package is latest (#{installedversion})"
        return installedversion
      when "<"
	# vim-lite-7.4.281                   <   needs updating (port has 7.4.295)
        if otherdata =~ /\(port has (\S+)\)/
          newversion = $1
          Puppet.debug "pkg version - Installed version needs updating to (#{newversion})"
          return newversion
        else
          Puppet.debug "pkg version - Unable to determine new version from (#{otherdata})"
          return installedversion
        end
      when "?", "!", "#"
        Puppet.debug "pkg version - Comparison Error reported from portversion (#{output})"
        return installedversion
      else
        Puppet.debug "pkg version - Unknown code from portversion output (#{output})"
        return installedversion
      end

    else
      # error: output not parsed correctly, error out with nil.
      # Seriously - this section should never be called in a perfect world.
      # as verification that the port is installed has already happened in query.
      if output =~ /^\*\* No matching package /
        raise Puppet::ExecutionFailure, "Could not find package #{@resource[:name]}"
      else
        # Any other error (dump output to log)
        raise Puppet::ExecutionFailure, "Unexpected output from portversion: #{output}"
      end

      # Just in case we still are running, return nil
      return nil
    end

    # At this point normal operation has finished and we shouldn't have been called.
    # Error out and let the admin deal with it.
    raise Puppet::Error, "pkg version - fatal error with pkg version: #{output}"
    return nil
  end

  ###### Query subcommand - return a hash of details if exists, or nil if it doesn't.
  # Used to make sure the package is installed

  def query
    Puppet.debug "pkg info - Called on #{@resource[:name]}"

    cmdline = ['info', '-e', @resource[:name]]
    begin
      output = pkg(*cmdline )
    rescue Puppet::ExecutionFailure
      ###raise Puppet::Error.new(output)
      Puppet.debug "package (#{@resource[:name]}) not installed"
      return nil
    end

    Puppet.debug "package (#{@resource[:name]}) is installed"

    ###cmdline = ['info', '-q', @resource[:name]]
    cmdline = ['query', '"%n-%v"', @resource[:name]]
    begin
      output = pkg(*cmdline)
    rescue Puppet::ExecutionFailure
      raise Puppet::Error.new(output)
    end

    # Check: if output isn't in the right format, return nil
    if output =~ /^(\S+)-([^-\s]+)/
      # Fill in the details
      hash = Hash.new
      hash[:portorigin] = self.name
      hash[:portname]   = $1
      hash[:ensure]     = $2

      # If more details are required, then we can do another pkg_info
      # query here and parse out that output and add to the hash
      # return the hash to the caller
      Puppet.debug( hash.to_yaml ) 
      return hash
    else
      Puppet.debug "pkg query - package (#{@resource[:name]}) not installed"
      return nil
    end
  end

  ####### Uninstall command

  def uninstall
    Puppet.debug "pkg delete - called on #{@resource[:name]}"
    # Get full package name from port origin to uninstall with
    cmdline = ['info','-e', @resource[:name]]
    begin
      output = pkg(*cmdline)
    rescue Puppet::ExecutionFailure
      #raise Puppet::Error.new(output)
      Puppet.debug "no such package found: #{@resource[:name]}"
      return nil
    end

      # output matches, so uninstall it
    Puppet.debug "uninstalling  #{@resource[:name]}"
    ###portuninstall $1
    cmdline = ['delete', '-y' , @resource[:name]]
    begin
      output = pkg(*cmdline)
    rescue Puppet::ExecutionFailure
      Puppet.debug "Cannot uninstall: #{@resource[:name]}: #{output}"	
      return nil
    end
    Puppet.debug output
    Puppet.debug "#{@resource[:name]} uninstalled"
  end

  ######## Update/upgrade command

  def update
    a_pty_device = PTY.open[1].path

    Puppet.debug "portupgrade - called on (#{@resource[:name]})"
    Puppet.debug "allocated pty device #{a_pty_device}"

    cmdline = ['info','-e', @resource[:name]]
    begin
      output = pkg(*cmdline)
    rescue Puppet::ExecutionFailure
      #raise Puppet::Error.new(output)
      Puppet.debug "no such package found: #{@resource[:name]}"
      return nil
    end

      # output matches, so upgrade the software
      cmdline = ['/usr/local/sbin/portupgrade','--batch','-v','-b','-M BATCH=yes', @resource[:name]]
      options = { :stdinfile => a_pty_device, :combine => :true, :raise => :true, :failonfail => :true }
      begin
        output = Puppet::Util::Execution.execute( cmdline, options ) 
      rescue Puppet::ExecutionFailure
        raise Puppet::Error.new(output)
      end
  end
end
